/*
 * Distributed under the OSI-approved Apache License, Version 2.0.  See
 * accompanying file Copyright.txt for details.
 *
 * ENGINE_H : deprecated wrapper - use bindings/CXX/adios2/cxx instead
 *
 * DEPRECATION WARNING: The bindings/CXX/adios2/cxx11 directory has been
 * renamed to bindings/CXX/adios2/cxx. Please update your includes to use
 * bindings/CXX/adios2/cxx instead of bindings/CXX/adios2/cxx11.
 */

#ifndef ADIOS2_BINDINGS_CXX_CXX11_ENGINE_H_
#define ADIOS2_BINDINGS_CXX_CXX11_ENGINE_H_

#if defined(_MSC_VER)
#pragma message("DEPRECATED: bindings/CXX/adios2/cxx11 has been renamed to bindings/CXX/adios2/cxx. Please update your includes.")
#else
#warning "DEPRECATED: bindings/CXX/adios2/cxx11 has been renamed to bindings/CXX/adios2/cxx. Please update your includes."
#endif

#include "../cxx/Engine.h"

#endif /* ADIOS2_BINDINGS_CXX_CXX11_ENGINE_H_ */
